/******************************************************************************
 *
 * $Workfile: uart.c $
 *
 * This module provides the interface routines for a bi-directional UART
 * running on two TPU channels.  One each for transmit and receive.
 *
 * $History: uart.c $
 *
 *****************************************************************************/
#include "tpu.h"
#include "uart.h"

#define ENABLE_INTS()    asm(" andi #~0x0700,sr")
#define DISABLE_INTS()   asm(" ori  #0x0700,sr")

#define XTAL_FREQ 16000000L

/* Define the channel numbers for the TPU serial interfaces.
 */

#define TPU_TXD   0
#define TPU_RXD   1

#define TCR1_FREQ (XTAL_FREQ/128)

static void uart_receive(void);
static void uart_transmit(void);

static int uart_tx_running;
static int uart_rx_overrun;
static int uart_tx_overrun;

#define RX_BUFFER_SIZE 32
static char rx_buffer[RX_BUFFER_SIZE];
static unsigned short rx_insert_ptr, rx_extract_ptr;

#define TX_BUFFER_SIZE 2048
static char tx_buffer[TX_BUFFER_SIZE];
static unsigned short tx_insert_ptr, tx_extract_ptr;

/* Define indices for accessing the parameter RAM */
enum UART_RAM {parity_temp = 0, match_rate, data_reg, data_size, bit_count, shift_reg};

/* Define the host service requests */
enum HSR_UART {receive = 2, transmit};

/******************************************************************************
 *
 * Function Name: void tpu_uart_init(unsigned int baud, int parity)
 *
 * Description:  
 *    Initialize the transmit and receive channels, setting the baud rate and
 *    parity. One stop bit will be presumed, because the UART cannot handle
 *    anything else on a transmit channel, so we'll keep the receive channel
 *    consistent with it.
 *
 *    The initialization sequence is prescribed by Motorola in TPU Programming 
 *    Note TPUPN07/D.
 *
 *    NOTE: The only TPU register which can be accessed on a byte basis is 
 *     CISR. ALL other registers must be accessed as words or long words! Be 
 *     very careful if you change the access methods for the registers!
 *
 * Calling Sequence: 
 *    unsigned int baud     desired baud rate 
 *    int parity            no_parity, even_parity, or odd_parity
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
void tpu_uart_init(unsigned int baud, int parity)
{
  /* Write channel priorities */
  tpu_cpr_set(TPU_TXD, 0);
  tpu_cpr_set(TPU_RXD, 0);

  /* Clear the channel function select field */
  tpu_cfs_set(TPU_TXD, 0);
  tpu_cfs_set(TPU_RXD, 0);

  /* Set the baud rate */
  PARAMETER_RAM(TPU_RXD)[match_rate] =
  PARAMETER_RAM(TPU_TXD)[match_rate] = (unsigned short)(TCR1_FREQ / baud);

  /* Set TDRE */
  PARAMETER_RAM(TPU_TXD)[data_reg] = 0x8000;

  /* Clear receiver and receive flags */
  PARAMETER_RAM(TPU_RXD)[data_reg] = 0x0000;

  /* Set the data size to 8 bit */
  PARAMETER_RAM(TPU_TXD)[data_size] =
  PARAMETER_RAM(TPU_RXD)[data_size] = 8;

  /* Write the channel function select field */
  tpu_cfs_set(TPU_TXD, TPU_UART);
  tpu_cfs_set(TPU_RXD, TPU_UART);

  /* Set the host sequence bits to select parity */
  tpu_hsq_set(TPU_TXD, parity);
  tpu_hsq_set(TPU_RXD, parity);

  /* Set up for servicing interrupts */
  tpu_int_setup((TPU_INT + TPU_RXD), (void *)uart_receive);
  tpu_int_setup((TPU_INT + TPU_TXD), (void *)uart_transmit);
  tpu_int_enable(TPU_RXD);

  /* Issue host service requests */
  tpu_hsr_set(TPU_TXD, transmit);
  tpu_hsr_set(TPU_RXD, receive);

  /* Clear any pending interrupts */
  tpu_int_clear(TPU_TXD);
  tpu_int_clear(TPU_RXD);

  /* Write channel priorities */
  tpu_cpr_set(TPU_TXD, TPU_PRIORITY_MEDIUM);
  tpu_cpr_set(TPU_RXD, TPU_PRIORITY_HIGH);

  /* "Empty" the character buffers. */
  rx_insert_ptr = rx_extract_ptr = 0;
  tx_insert_ptr = tx_extract_ptr = 0;

  // clear tx_running flags & error counters
  uart_tx_running = 0;
  uart_rx_overrun = 0;
  uart_tx_overrun = 0;
}


/******************************************************************************
 *
 * Function Name: int tpu_uart_putc(char character)
 *
 * Description:  
 *    Send a character via a transmit channel.
 *
 * Calling Sequence: 
 *    char character        character to transmit
 *
 * Returns:
 *     0 on success
 *    -1 on error (queue full)
 *
 *****************************************************************************/
int tpu_uart_putc(char character)
{
  int temp;

  temp = (tx_insert_ptr + 1) % TX_BUFFER_SIZE;

  if (temp == tx_extract_ptr)
    {
    uart_tx_overrun++;
    return -1;                          /* no room */
    }

  tx_buffer[tx_insert_ptr] = character;
  tx_insert_ptr = temp;

  DISABLE_INTS();

  if (!uart_tx_running)
    {
    PARAMETER_RAM(TPU_TXD)[data_reg] = 0x8000; /* Set TDRE */
    tpu_int_enable(TPU_TXD);           /* enable interrupts */
    uart_tx_running = 1;
    }

  ENABLE_INTS();
  return 0;
}


/******************************************************************************
 *
 * Function Name: size_t tpu_uart_space(void)
 *
 * Description:  
 *    get available space in uart transmit queue
 *
 * Calling Sequence: 
 *    void
 *
 * Returns:
 *    number of free bytes in queue
 *
 *****************************************************************************/
size_t tpu_uart_space(void)
{
  int space;

  if ((space = (tx_extract_ptr - tx_insert_ptr)) <= 0)
    space += TX_BUFFER_SIZE;

  return space - 1;
}


/******************************************************************************
 *
 * Function Name: char *tpu_uart_puts(char *string)
 *
 * Description:  
 *    Write NULL terminated 'string' to the uart output queue.
 *
 * Calling Sequence: 
 *    char *string      string to transmit
 *
 * Returns:
 *    pointer to the next character to be written.  (\0 if full string is 
 *    written)
 *
 *****************************************************************************/
char *tpu_uart_puts(char *string)
{
  register char c;

  while ((c = *string) && (tpu_uart_putc(c) == 0))
    string++;

  return string;
}

/******************************************************************************
 *
 * Function Name: void tpu_uart_flush_tx(void)
 *
 * Description:  
 *    flush the UART transmit buffer
 *
 * Calling Sequence: 
 *    void
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
void tpu_uart_flush_tx(void)
{
  DISABLE_INTS();

  // "Empty" the character buffer.
  tx_insert_ptr = tx_extract_ptr = 0;
  ENABLE_INTS();
}


/******************************************************************************
 *
 * Function Name: int tpu_uart_write(char *buffer, size_t count)
 *
 * Description:  
 *    Write 'count' characters from 'buffer' to the uart output queue.
 *    NOTE: if insufficient room, no characters are written.
 *
 * Calling Sequence: 
 *    char *buffer      pointer to characters to write
 *    size_t count      number of characters to write
 *
 * Returns:
 *     0 on success 
 *    -1 if insufficient room 
 *    -2 on error
 *
 *****************************************************************************/
int tpu_uart_write(char *buffer, size_t count)
{
  if (count > tpu_uart_space())
    {
    uart_tx_overrun++;
    return -1;                          /* no room */
    }

  while (count && (tpu_uart_putc(*buffer++) == 0))
    count--;

  return (count ? -2 : 0);
}


/******************************************************************************
 *
 * Function Name: int tpu_uart_getc(void)
 *
 * Description:  
 *    Get a character via a receive channel.
 *
 * Calling Sequence: 
 *    void
 *
 * Returns:
 *    character on success 
 *    -1 if no character is available
 *
 *****************************************************************************/
int tpu_uart_getc(void)
{
  unsigned char character;

  if (rx_insert_ptr == rx_extract_ptr)  /* check if character is available */
    return -1;

  character = rx_buffer[rx_extract_ptr++]; /* Get character, bump pointer */
  rx_extract_ptr %= RX_BUFFER_SIZE;     /* Limit the pointer */
  return character;
}


/******************************************************************************
 *
 * Function Name: static void uart_transmit(void)
 *
 * Description:  
 *    UART transmit interrupt handler - if there is a character in the queue,
 *    write it, else disable the transmit interrupt.
 *
 * Calling Sequence: 
 *    void
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
#pragma interrupt()
static void uart_transmit(void)
{
  unsigned char ch;

  if (tx_insert_ptr != tx_extract_ptr)
    {
    /* This is so that we don't require a subroutine call in the interrupt
     * handler to clear the ISR bit.
     * NOTE: BOTH OPERANDS ARE DEPENDENT UPON CHANNEL NUMBER!!!
     */
    asm(" bclr.b #0,$FFFE21");

    ch = tx_buffer[tx_extract_ptr++];
    tx_extract_ptr %= TX_BUFFER_SIZE;
    PARAMETER_RAM(TPU_TXD)[data_reg] = (unsigned short)ch;
    }
  else
    {
    tpu_int_disable(TPU_TXD);
    uart_tx_running = 0;
    }
}


/******************************************************************************
 *
 * Function Name: static void uart_receive(void)
 *
 * Description:  
 *    UART receiver interrupt handler - if room in the queue, get the
 *    character and put it in the queue, else increment the overrun counter.
 *
 * Calling Sequence: 
 *    void
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
#pragma interrupt()
static void uart_receive(void)
{
  int temp;

  /* This is so that we don't require a subroutine call in the interrupt
   * handler to clear the ISR bit.
   * NOTE: BOTH OPERANDS ARE DEPENDENT UPON CHANNEL NUMBER!!!
   */
  asm(" bclr.b #1,$FFFE21");

  temp = (rx_insert_ptr + 1) % RX_BUFFER_SIZE;

  if (temp != rx_extract_ptr)
    {
    rx_buffer[rx_insert_ptr] = (char)PARAMETER_RAM(TPU_RXD)[data_reg];
    PARAMETER_RAM(TPU_RXD)[data_reg] = 0;
    rx_insert_ptr = temp;
    }
  else
    uart_rx_overrun++;
}
